
import {vec3, mat4} from "gl-matrix";
import { MouseEventHandler, MouseButtons } from "./MouseEventHandler.js"
import { KeyboardEventHandler } from "./KeyboardEventHandler.js"

class Camera {

    //protected 
    constructor(name) {
        this.name_ = name;
        this.view_ = mat4.create();
        this.projection_ = mat4.create();
        this.viewProjection_ = mat4.create();

        this.nearPlane_ = 0.1;
        this.farPlane_ = 512.0;
        this.fov_ = 60.0;
        this.viewPortWidth_ = 512.0;
        this.viewPortHeight_ = 512.0;
        this.aspectRatio_ = this.viewPortWidth_ / this.viewPortHeight_;

        // Free camera
        this.yawPitchRoll_ = vec3.fromValues(0.0, 0.0, 0.0);
        this.cameraPosition_ = vec3.fromValues(0.0, 0.0, 0.0);
        this.cameraRotationSpeed_ = 10.0;
        this.cameraMovementSpeed_ = 0.1;

        this.cameraOnb = [vec3.fromValues(1.0, 0.0, 0.0), vec3.fromValues(0.0, 1.0, 0.0), vec3.fromValues(0.0, 0.0, 1.0)];
    }

    setViewport(width, height) {
        this.viewPortWidth_ = width;
        this.viewPortHeight_ = height;
        this.aspectRatio_ = this.viewPortWidth_ / this.viewPortHeight_;
    }

    setLookAt(pos, target, up) {
        this.cameraPosition_ = pos;
        mat4.lookAt(this.view_, pos, target, up);
    }

    setPerspective(fovInDeg) {
        this.fov_ = fovInDeg;
        mat4.perspective(this.projection_, this.fov_ * (Math.PI / 180.0), this.aspectRatio_, this.nearPlane_, this.farPlane_);
    }

    getView() { return this.view_; }
    getProjection() { return this.projection_; }
    getViewProjection() {
        mat4.multiply(this.viewProjection_, this.projection_, this.view_);
        return this.viewProjection_;
    }

    getRight() { return this.cameraOnb[0]; }
    getUp() { return this.cameraOnb[1]; }
    getDir() { return this.cameraOnb[2]; }

    getPosition() { return this.cameraPosition_;}

    update(deltaTime, mouseEventHandler, keyboardEventHandler) {

        let isUpdated = false;

        // update free camera
        const yawUpdate = -(mouseEventHandler.getMouseDeltaX() * this.cameraRotationSpeed_ * deltaTime);
        const pitchUpdate = (mouseEventHandler.getMouseDeltaY() * this.cameraRotationSpeed_ * deltaTime) / this.aspectRatio_;
        const movementDelta = this.cameraMovementSpeed_ * deltaTime;

        if (mouseEventHandler.isDown(MouseButtons.LeftButton)) {
            this.yawPitchRoll_[0] += yawUpdate;
            this.yawPitchRoll_[1] += pitchUpdate;
            isUpdated = true;
          //  console.log('updated', yawUpdate);
        }

        if (keyboardEventHandler.isDown('w')) {
            let v = vec3.fromValues(0.0, 0.0, 0.0);
            vec3.scale(v, this.getDir(), movementDelta);
            vec3.add(this.cameraPosition_, this.cameraPosition_, v);
        }
        if (keyboardEventHandler.isDown('s')) {
            let v = vec3.fromValues(0.0, 0.0, 0.0);
            vec3.scale(v, this.getDir(), -movementDelta);
            vec3.add(this.cameraPosition_, this.cameraPosition_, v);
        }
        if (keyboardEventHandler.isDown('a')) {
            let v = vec3.fromValues(0.0, 0.0, 0.0);
            vec3.scale(v, this.getRight(), -movementDelta);
            vec3.add(this.cameraPosition_, this.cameraPosition_, v);
        }
        if (keyboardEventHandler.isDown('d')) {
            let v = vec3.fromValues(0.0, 0.0, 0.0);
            vec3.scale(v, this.getRight(), movementDelta);
            vec3.add(this.cameraPosition_, this.cameraPosition_, v);
        }
        if (keyboardEventHandler.isDown('q')) {
            let v = vec3.fromValues(0.0, 0.0, 0.0);
            vec3.scale(v, this.getUp(), -movementDelta);
            vec3.add(this.cameraPosition_, this.cameraPosition_, v);
        }
        if (keyboardEventHandler.isDown('e')) {
            let v = vec3.fromValues(0.0, 0.0, 0.0);
            vec3.scale(v, this.getUp(), movementDelta);
            vec3.add(this.cameraPosition_, this.cameraPosition_, v);
        }

        let cameraOrientation_ = mat4.create();
        let cameraOrientationInv_ = mat4.create();
        let cameraTranslation_ = mat4.create();
        let cameraView_ = mat4.create();

        let rotX = mat4.create();
        let rotY = mat4.create();

        mat4.fromTranslation(cameraTranslation_, vec3.fromValues(-this.cameraPosition_[0], -this.cameraPosition_[1], -this.cameraPosition_[2]));
        mat4.fromXRotation(rotX,  this.yawPitchRoll_[1]);
        mat4.fromYRotation(rotY,  this.yawPitchRoll_[0]);
        mat4.multiply(cameraOrientation_, rotY, rotX); 
        mat4.transpose(cameraOrientationInv_, cameraOrientation_);
        mat4.multiply(cameraView_, cameraOrientationInv_, cameraTranslation_);
        this.view_ = cameraView_;

        this.cameraOnb[0][0] = cameraOrientation_[0];
        this.cameraOnb[0][1] = cameraOrientation_[1];
        this.cameraOnb[0][2] = cameraOrientation_[2];

        this.cameraOnb[1][0] = cameraOrientation_[4];
        this.cameraOnb[1][1] = cameraOrientation_[5];
        this.cameraOnb[1][2] = cameraOrientation_[6];

        this.cameraOnb[2][0] = -cameraOrientation_[8];
        this.cameraOnb[2][1] = -cameraOrientation_[9];
        this.cameraOnb[2][2] = -cameraOrientation_[10];

        return isUpdated;

    }
}

export { Camera };
