
// External //////////////////////////////////////////////////////////////////
import { vec3, vec4, mat4 } from "gl-matrix";
import dat from "dat.gui";

// wtek
import { wtShader } from "wtek/wtShader.js"
import { wtSwapChain } from "wtek/wtSwapChain.js"
import { wtBuffer } from "wtek/wtBuffer.js"
import { wtBindGroupLayout } from "wtek/wtBindGroupLayout.js"
import { wtVertexDescriptor } from "wtek/wtVertexDescriptor.js"
import { VertexType } from "wtek/wtVertexDescriptor.js"
import { wtRenderPipeline, GpuPrimTopology } from "wtek/wtRenderPipeline.js"
import { wtTexture } from "wtek/wtTexture.js"
import { wtComputePipeline } from "wtek/wtComputePipeline.js"
import { wtPrimRender } from "wtek/wtPrimRender.js"
import { Vec4Colors } from "wtek/wContants.js"

// wfun
import { Camera } from "./wfun/Camera.js"
import { ModelLoader } from "./wfun/ModelLoader.js"
import { MouseEventHandler } from "./wfun/MouseEventHandler.js"
import { KeyboardEventHandler } from "./wfun/KeyboardEventHandler.js"

// shader import test
import vertexShaderGLSL from "./glsl/vertexShader.glsl";
import fragmentShaderGLSL from "./glsl/fragShader.glsl";
import pointShaderSrcVS from "./glsl/primPointVS.glsl";
import pointShaderSrcFS from "./glsl/primPointFS.glsl";
import pixelShaderSrcVS from "./glsl/primPixelBufferVS.glsl";
import pixelShaderSrcFS from "./glsl/primPixelBufferFS.glsl";

// Uniform buffer data///////////////////////////////////////////////////////////
class uniformData {
    constructor() {
        this.projectionMatrix_ = mat4.create();
        this.modelMatrix_ = mat4.create();
        this.imageDimension_ = vec4.fromValues(0.0, 0.0, 0.0, 0.0);
        this.cameraPosition_ = vec4.fromValues(0.0, 0.0, 0.0, 0.0);
        this.floatArray_ = new Float32Array(32 + 4 + 4);
    }

    toArray() {
        let a0 = Float32Array.from(this.projectionMatrix_);
        let a1 = Float32Array.from(this.modelMatrix_);

        for (let i = 0; i < 16; i++) {
            this.floatArray_[i] = a0[i];
            this.floatArray_[i + 16] = a1[i];
        }

        for (let i = 0; i < 4; i++) {
            this.floatArray_[32 + i] = this.imageDimension_[i];
            this.floatArray_[36 + i] = this.cameraPosition_[i];
        }
        return this.floatArray_;
    }

    updateCameraPosition(p) {
        this.cameraPosition_[0] = p[0];
        this.cameraPosition_[1] = p[1];
        this.cameraPosition_[2] = p[2];
    }

    updateAnimationTimer(time) {
        this.imageDimension_[3] += time;
    }

    setImageDimenion(width, height) {
        this.imageDimension_[0] = width;
        this.imageDimension_[1] = height;
    }

    getSizeInBytes() {
        return this.floatArray_.length * Float32Array.BYTES_PER_ELEMENT;
    }
}

// App /////////////////////////////////////////////////////////////////////////
class wtApp {

    constructor(gpuContext, glslang, canvas) {
        this.gpuContext_ = gpuContext;
        this.glslang_ = glslang;
        this.canvas_ = canvas;
        this.camera_ = new Camera('mainCamera');
        this.frameCount_ = 0;
        this.lastFps_ = 0;
        this.msaaSampleCount_ = 4;
        this.swapChainFormat_ = "bgra8unorm";

        //event handlers
        this.keyboardEventHandler_ = new KeyboardEventHandler();
        this.mouseEventHander_ = new MouseEventHandler();

        // todo in camera
        this.projectionMatrix = mat4.create();
        this.modelMatrix = mat4.create();

        this.cameraDir = vec3.fromValues(0.0, 1.0, 0.0);
        this.cameraUp = vec3.fromValues(0.0, 1.0, 0.0);
        this.cameraPosition = vec3.fromValues(0.0, 0.3, 2.0);

        this.camera_.setViewport(canvas.width, canvas.height);
        this.camera_.setPerspective(60.0);
        this.camera_.setLookAt(this.cameraPosition, vec3.fromValues(0.5, 3.0, 0.0), this.cameraUp);
        this.camera_.cameraMovementSpeed_ = 0.5;

        this.palette = {
            color1: [255, 0, 0], // RGB array, // CSS string
            color2: [0, 255, 0], // RGB array
            color3: [0, 0, 255], // RGB with alpha
        };


        // dat gui
        this.browerInfo_ = {
            browser: "unknown",
            winner: true
        };

        this.slider_ = {
            name: "Slider",
            scale: 0.15,
            maxSize: 4.0
        }

        this.appInfo_ = {
            appName: "WebGpu Primitive Render",
            appDeveloper: "Peter Trier & Thomas Kim Kjeldsen",
            fps: 0,
        }

        this.debugGui_ = new dat.GUI({ name: 'Debug input' });
        this.debugGui_.width = 370.0;
        this.debugGui_.add(this.appInfo_, 'appName');
        this.debugGui_.add(this.appInfo_, 'appDeveloper');
        this.debugGui_.addColor(this.palette, 'color1');
        this.debugGui_.addColor(this.palette, 'color2');
        this.debugGui_.addColor(this.palette, 'color3');
        this.debugGui_.add(this.slider_, "scale").min(0.1).max(0.8);

        let fpsController = this.debugGui_.add(this.appInfo_, 'fps');

        setInterval(() => {
            fpsController.setValue(this.appInfo_.fps)
        }, 1000);

    }

    msaaEnabled() { return this.msaaSampleCount_ > 1; }
    getKeyboardEventHandler() { return this.keyboardEventHandler_; }
    getMouseEventHandler() { return this.mouseEventHander_; }

    async init(browserId) {
        // create the swap chain
        let ctx = null
        this.browerInfo_.browser = browserId;
        this.debugGui_.add(this.browerInfo_, "browser");

        if (browserId === 'Chrome' || browserId === 'Firefox') {
            ctx = this.canvas_.getContext('gpupresent');
        }
        else {
            ctx = this.canvas_.getContext('gpupresent');
        }

        if (ctx === undefined) {
            alert('Unable to create gpu canvas context')
        }

        this.swapChain_ = new wtSwapChain('swapChain', this.gpuContext_);
        this.swapChain_.createSwapChain(this.swapChainFormat_, ctx);

        this.depthStencil_ = new wtTexture('depthStencilTexture', this.gpuContext_, GPUTextureUsage.OUTPUT_ATTACHMENT);
        this.depthStencil_.create2D(this.gpuContext_.getWidth(), this.gpuContext_.getHeight(), 'depth24plus-stencil8', this.msaaSampleCount_);

        if (this.msaaEnabled()) {
            this.texMsaa_ = new wtTexture('multiSampledTexture', this.gpuContext_, GPUTextureUsage.OUTPUT_ATTACHMENT);
            this.texMsaa_.create2D(this.gpuContext_.getWidth(), this.gpuContext_.getHeight(), this.swapChainFormat_, this.msaaSampleCount_);
            this.renderAttachment_ = this.texMsaa_.createView();
            this.depthStencilAttachment_ = this.depthStencil_.createView();
        }
        else {
            this.depthStencilAttachment_ = this.depthStencil_.createView();
        }

        await this.loadTextureData();
        await this.loadModelData();

        this.buildStorageBuffers();
        this.buildUniformBuffers();
        this.buildUniformBindDescriptors();
        this.buildVertexBuffers();
        this.buildRenderPipelines();
    //    this.buildComputePipelines();
    }

    async loadTextureData() {

    }

    async loadModelData() {
        const response = await fetch('./assets/bunny1000.obj');
        const objModelText = await response.text();
        this.modelLoader_ = new ModelLoader();
        this.modelLoader_.parseObjFile(objModelText);
    }

    buildVertexBuffers() {
        this.vertexBuffer_ = new wtBuffer('vertBuffer', this.gpuContext_);
        this.vertexBuffer_.createVertexBufferFromData(new Float32Array(
            [
                -1.0, -1.0, 0.0, 1.0,
                0.0, 1.0, 0.0, 1.0,
                1.0, -1.0, 0.0, 1.0
            ]
        ));
    }

    buildStorageBuffers() {
        let pixelBufferSize = this.gpuContext_.getWidth() * this.gpuContext_.getHeight() * 4 * Float32Array.BYTES_PER_ELEMENT;
        this.storagePixelBuffer_ = new wtBuffer('pixelBuffer', this.gpuContext_);
        this.storagePixelBuffer_.createStorageBuffer(pixelBufferSize);
    }

    buildUniformBuffers() {
        let layOutEntry0 = {
            binding: 0,
            visibility: GPUShaderStage.VERTEX | GPUShaderStage.FRAGMENT,
            type: "uniform-buffer"
        };

        this.uniformData_ = new uniformData();
        this.uniformData_.setImageDimenion(this.gpuContext_.getWidth(), this.gpuContext_.getHeight());

        this.uniformBindGroupLayout_ = new wtBindGroupLayout('uniformGroupLayout0', this.gpuContext_);
        this.uniformBindGroupLayout_.addLayoutBinding(layOutEntry0);
        this.uniformBindGroupLayout_.create();
        this.uniformBuffer_ = new wtBuffer('uniformBuffer', this.gpuContext_);
        this.uniformBuffer_.createUniformBuffer(this.uniformData_.getSizeInBytes());
    }

    buildUniformBindDescriptors() {
        const uniformBindGroupDescriptor = {
            layout: this.uniformBindGroupLayout_.getBindGroupLayout(),
            entries: [
                {
                    binding: 0,
                    resource: {
                        buffer: this.uniformBuffer_.getBuffer(),
                        offset: 0,
                        size: this.uniformBuffer_.getSizeInBytes()
                    }
                }
            ]
        };
        this.uniformBindGroup_ = this.gpuContext_.getDevice().createBindGroup(uniformBindGroupDescriptor);
    }

    buildRenderPipelines() {
        this.shader_ = new wtShader('shaders', this.gpuContext_);
        let vShaderModule = this.shader_.createVertexShaderModule(this.glslang_, vertexShaderGLSL);
        let fShaderModule = this.shader_.createFragmentShaderModule(this.glslang_, fragmentShaderGLSL);

        let vertexDescriptor = new wtVertexDescriptor(VertexType.VT_Vertex4);
        this.renderPipeline_ = new wtRenderPipeline('renderPipeline', this.gpuContext_);
        this.renderPipeline_.setVertexModule(vShaderModule);
        this.renderPipeline_.setFragmentModule(fShaderModule);
        this.renderPipeline_.setBindGroupLayouts(this.uniformBindGroupLayout_.getBindGroupLayout());
        this.renderPipeline_.setPrimTopology(GpuPrimTopology.TriangleList);
        this.renderPipeline_.setVertexBufferState(vertexDescriptor.getVertexState());
        this.renderPipeline_.setSampleCount(this.msaaSampleCount_);
        this.renderPipeline_.create();

        // prim render pipeline creation
        this.primRender_ = new wtPrimRender('primRender', this.gpuContext_);
        let pointVS = this.shader_.createVertexShaderModule(this.glslang_, pointShaderSrcVS);
        let pointFS = this.shader_.createFragmentShaderModule(this.glslang_, pointShaderSrcFS);
        this.primRender_.createPipelines(pointVS, pointFS, this.msaaSampleCount_);

        let pixelVS = this.shader_.createVertexShaderModule(this.glslang_, pixelShaderSrcVS);
        let pixelFS = this.shader_.createFragmentShaderModule(this.glslang_, pixelShaderSrcFS);
        this.primRender_.createPixelPipeline(pixelVS, pixelFS, this.msaaSampleCount_);
        this.primRender_.setPixelStorageBuffer(this.storagePixelBuffer_);
    }

    buildComputePipelines() {
        let computeShaderModule = this.shader_.createComputeShaderModule(this.glslang_, rayMarchSrcCS);
        this.computePipeline_ = new wtComputePipeline('computePipeline', this.gpuContext_);
        this.computePipeline_.create(computeShaderModule);

        let layOutEntry0 = {
            binding: 0,
            visibility: GPUShaderStage.COMPUTE,
            type: "uniform-buffer"
        };

        let layOutEntry1 = {
            binding: 1,
            visibility: GPUShaderStage.COMPUTE,
            type: "storage-buffer"
        };

        const uniformBindGroupDescriptor = {
            layout: this.computePipeline_.getPipeline().getBindGroupLayout(0),
            entries: [
                {
                    binding: 0,
                    resource: {
                        buffer: this.uniformBuffer_.getBuffer(),
                        offset: 0,
                        size: this.uniformBuffer_.getSizeInBytes()
                    }
                },
                {
                    binding: 1,
                    resource: {
                        buffer: this.storagePixelBuffer_.getBuffer(),
                        offset: 0,
                        size: this.storagePixelBuffer_.getSizeInBytes()
                    }
                },
            ]
        };

        this.computeBindGroup_ = this.gpuContext_.getDevice().createBindGroup(uniformBindGroupDescriptor);
    }

    update(deltaTime) {
        this.camera_.update(deltaTime, this.getMouseEventHandler(), this.getKeyboardEventHandler());
        this.getMouseEventHandler().flush();
        this.uniformData_.updateAnimationTimer(deltaTime);
        this.uniformData_.updateCameraPosition(this.camera_.getPosition());
    }

    render(timestamp) {
        this.updateFps();

        const commandEncoder = this.gpuContext_.getDevice().createCommandEncoder();
        const textureView = this.swapChain_.getSwapChain().getCurrentTexture().createView();
        let renderPassDescriptor;

        if (this.msaaEnabled()) {
            renderPassDescriptor = {
                colorAttachments: [{
                    attachment: this.renderAttachment_,
                    resolveTarget: textureView,
                    loadValue: { r: 0.3, g: 0.3, b: 0.4, a: 1.0 },
                }],
                depthStencilAttachment: {
                    attachment: this.depthStencil_.createView(),
                    depthLoadValue: 1.0,
                    depthStoreOp: "store",
                    stencilLoadValue: 0,
                    stencilStoreOp: "store",
                }
            };

        }
        else {
            renderPassDescriptor = {
                colorAttachments: [{
                    attachment: textureView,
                    loadValue: { r: 0.4, g: 0.4, b: 0.4, a: 1.0 },
                }],
                depthStencilAttachment: {
                    attachment: this.depthStencilAttachment_,
                    depthLoadValue: 1.0,
                    depthStoreOp: "store",
                    stencilLoadValue: 0,
                    stencilStoreOp: "store",
                }
            };
        }


        this.projectionMatrix = this.camera_.getViewProjection();
        let animSpeed = timestamp / 2000.0;
        mat4.identity(this.modelMatrix);
        mat4.translate(this.modelMatrix, this.modelMatrix, [Math.sin(animSpeed), Math.cos(animSpeed), 0.0]);
        mat4.rotateX(this.modelMatrix, this.modelMatrix, animSpeed);
        mat4.rotateY(this.modelMatrix, this.modelMatrix, animSpeed);
        mat4.rotateZ(this.modelMatrix, this.modelMatrix, animSpeed);
       
        this.primRender_.clearPrimitive();
        this.primRender_.addGridXZ(16.0, 0.1);
        this.primRender_.addOrigin(vec4.fromValues(0.0, 0.0, 0.0, 1.0), 0.2);
     
        // points
        let objVerts = this.modelLoader_.getVertexData();
        let scaler =  this.slider_.scale;

        for (let i = 0; i < objVerts.length; i++) {
            let val = objVerts[i];
            let p = vec4.fromValues(val[0] * scaler, val[1] * scaler, val[2] * scaler, 1.0);
            this.primRender_.addPoint(p, Vec4Colors.LightGrey);
        }

        // triangle test
        let p0 = vec4.fromValues(0.0,0.0,0.0,0.0);
        let p1 = vec4.fromValues(0.0,0.0,0.0,0.0);
        let p2 = vec4.fromValues(0.0,0.0,0.0,0.0);

        vec4.transformMat4(p0,  vec4.fromValues(-2.0, 0.0, 0.0, 1.0), this.modelMatrix);
        vec4.transformMat4(p1,  vec4.fromValues(-1.0, 0.0, -0.5, 1.0), this.modelMatrix);
        vec4.transformMat4(p2,  vec4.fromValues(-1.0, 1.0, -0.6, 1.0), this.modelMatrix);

        let c0 = vec4.fromValues(this.palette.color1[0] / 255.0, this.palette.color1[1] / 255.0, this.palette.color1[2] / 255.0, 1.0);
        let c1 = vec4.fromValues(this.palette.color2[0] / 255.0, this.palette.color2[1] / 255.0, this.palette.color2[2] / 255.0, 1.0);
        let c2 = vec4.fromValues(this.palette.color3[0] / 255.0, this.palette.color3[1] / 255.0, this.palette.color3[2] / 255.0, 1.0);

        this.primRender_.addTriangle(p0, c0, p1, c1, p2, c2);
        this.primRender_.update();
        this.primRender_.updateUniforms(this.camera_.getViewProjection());

        // update uniforms
        this.uniformData_.projectionMatrix_ = this.projectionMatrix;
        this.uniformData_.modelMatrix_ = this.modelMatrix;
        let uniformArray = this.uniformData_.toArray();
        this.gpuContext_.getDevice().defaultQueue.writeBuffer(this.uniformBuffer_.getBuffer(), 0, uniformArray.buffer, 0, this.uniformBuffer_.getSizeInBytes());

        {
            const passEncoder = commandEncoder.beginRenderPass(renderPassDescriptor);
            this.primRender_.render(passEncoder);
            passEncoder.endPass();
        }

        this.gpuContext_.getDevice().defaultQueue.submit([commandEncoder.finish()]);
    }

    updateFps() {
        this.frameCount_++;
        let timeSinceLastFpsMeasure = (performance.now() - this.lastFps_) / 1000.0;
        if (timeSinceLastFpsMeasure > 1.0) {
            let fps = this.frameCount_ / timeSinceLastFpsMeasure;
            this.lastFps_ = performance.now();
            this.frameCount_ = 0;
            this.appInfo_.fps = Math.round(fps);
        }
    }

    resize(width, height) {
    }

}

export { wtApp };
